# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/utils/00_basic.ipynb.

# %% auto 0
__all__ = ['load_catalog', 'load_params', 'DF_TYPE', 'pmap', 'DataConfig', 'filter_tranges', 'filter_tranges_df', 'pl_norm',
           'partition_data_by_ts', 'partition_data_by_year', 'partition_data_by_year_month', 'partition_data_by_time',
           'concat_df', 'concat_partitions', 'format_timedelta', 'resample', 'df2ts', 'calc_vec_mag', 'check_fgm']

# %% ../../notebooks/utils/00_basic.ipynb 1
from .. import ROOT_DIR
from .kedro import load_context

from functools import partial

from typing import overload

# %% ../../notebooks/utils/00_basic.ipynb 2
load_catalog = partial(load_context, project_path=ROOT_DIR, catalog_only=True)
load_params = partial(load_context, project_path=ROOT_DIR, params_only=True)

# %% ../../notebooks/utils/00_basic.ipynb 4
import polars as pl
import polars.selectors as cs

import pandas as pd
import xarray as xr

import pandas
import numpy as np
from xarray_einstats import linalg

from datetime import timedelta

from loguru import logger

from xarray import DataArray
from typing import Union, Collection, Callable, Optional, Tuple
from typing import Any, Dict


# %% ../../notebooks/utils/00_basic.ipynb 5
from pipe import select
from fastcore.utils import partial

# %% ../../notebooks/utils/00_basic.ipynb 6
def pmap(func, *args, **kwargs):
    """
    map with `partial`
    """
    return select(partial(func, *args, **kwargs))

# %% ../../notebooks/utils/00_basic.ipynb 9
from pydantic import BaseModel
from datetime import datetime, timedelta
from pandas import Timedelta

class DataConfig(BaseModel):
    sat_id: str = None
    start: datetime = None
    end: datetime = None
    ts: timedelta = None
    coord: str = None

# %% ../../notebooks/utils/00_basic.ipynb 11
from fastcore.utils import patch

# %% ../../notebooks/utils/00_basic.ipynb 12
def filter_tranges(time: pl.Series, tranges: Tuple[list, list]):
    """
    - Filter data by time ranges, return the indices of the time that are in the time ranges (left inclusive, right exclusive)
    """

    starts = tranges[0]
    ends = tranges[1]

    start_indices = time.search_sorted(starts)
    end_indices = time.search_sorted(ends)

    return np.concatenate(
        [
            np.arange(start_index, end_index)
            for start_index, end_index in zip(start_indices, end_indices)
        ]
    )

def filter_tranges_df(df: pl.DataFrame, tranges: Tuple[list, list], time_col: str = "time"):
    """
    - Filter data by time ranges
    """

    time = df[time_col]
    filtered_indices = filter_tranges(time, tranges)
    return df[filtered_indices]

# %% ../../notebooks/utils/00_basic.ipynb 13
@patch
def plot(self:pl.DataFrame, *args, **kwargs):
    return self.to_pandas().plot(*args, **kwargs)

# %% ../../notebooks/utils/00_basic.ipynb 14
def _expand_selectors(items: Any, *more_items: Any) -> list[Any]:
    """
    See `_expand_selectors` in `polars`.
    """
    expanded: list[Any] = []
    for item in (
        *(
            items
            if isinstance(items, Collection) and not isinstance(items, str)
            else [items]
        ),
        *more_items,
    ):
        expanded.append(item)
    return expanded

def pl_norm(columns, *more_columns) -> pl.Expr:
    """
    Computes the square root of the sum of squares for the given columns.

    Args:
    *columns (str): Names of the columns.

    Returns:
    pl.Expr: Expression representing the square root of the sum of squares.
    """
    all_columns = _expand_selectors(columns, *more_columns)
    squares = [pl.col(column).pow(2) for column in all_columns]

    return sum(squares).sqrt()

# %% ../../notebooks/utils/00_basic.ipynb 16
def partition_data_by_ts(df: pl.DataFrame, ts: timedelta) -> Dict[str, pl.DataFrame]:
    """Partition the dataset by time

    Args:
        df: Input DataFrame.
        ts: Time interval.

    Returns:
        Partitioned DataFrame.
    """
    return df.with_columns(
        key=pl.col("time").dt.truncate(ts).cast(pl.Utf8)
    ).partition_by("key", include_key=False, as_dict=True)


def partition_data_by_year(df: pl.DataFrame) -> Dict[str, pl.DataFrame]:
    """Partition the dataset by year

    Args:
        df: Input DataFrame.

    Returns:
        Partitioned DataFrame.
    """
    return df.with_columns(year=pl.col("time").dt.year().cast(pl.Utf8)).partition_by(
        "year", include_key=False, as_dict=True
    )


def partition_data_by_year_month(df: pl.DataFrame) -> Dict[str, pl.DataFrame]:
    """Partition the dataset by year

    Args:
        df: Input DataFrame.

    Returns:
        Partitioned DataFrame.
    """
    return df.with_columns(
        year_month=pl.col("time").dt.year().cast(pl.Utf8)
        + "_"
        + pl.col("time").dt.month().cast(pl.Utf8).str.zfill(2),
    ).partition_by("year_month", include_key=False, as_dict=True)
    
def partition_data_by_time(df: pl.LazyFrame | pl.DataFrame, method) -> Dict[str, pl.DataFrame]:
    """Partition the dataset by time

    Args:
        df: Input DataFrame.
        method: The method to partition the data.

    Returns:
        Partitioned DataFrame.
    """
    if isinstance(df, pl.LazyFrame):
        df = df.collect()
    
    if method == "year":
        return partition_data_by_year(df)
    elif method == "year_month":
        return partition_data_by_year_month(df)
    else:
        ts = pd.Timedelta(method)
        return partition_data_by_ts(df, ts)

# %% ../../notebooks/utils/00_basic.ipynb 17
DF_TYPE = Union[pl.DataFrame, pl.LazyFrame, pd.DataFrame]
def concat_df(dfs: list[DF_TYPE]) -> DF_TYPE:
    """Concatenate a list of DataFrames into one DataFrame.
    """
    
    match type(dfs[0]):
        case pl.DataFrame | pl.LazyFrame:
            concat_func = pl.concat
        case pandas.DataFrame:
            concat_func = pandas.concat
        case _:
            raise ValueError(f"Unsupported DataFrame type: {type(dfs[0])}")
    
    return concat_func(dfs)
                     
def concat_partitions(partitioned_input: Dict[str, Callable]):
    """Concatenate input partitions into one DataFrame.

    Args:
        partitioned_input: A dictionary with partition ids as keys and load functions as values.
    """
    partitions_data = [
        partition_load_func() for partition_load_func in partitioned_input.values()
    ]  # load the actual partition data
    
    result = concat_df(partitions_data)
    return result

# %% ../../notebooks/utils/00_basic.ipynb 19
def format_timedelta(time):
    """Format timedelta to `timedelta`"""
    if isinstance(time, timedelta):
        return time
    elif isinstance(time, str):
        return pd.Timedelta(time)
    elif isinstance(time, int):
        return pd.Timedelta(seconds=time)
    else:
        raise TypeError(f"Unsupported type: {type(time)}")

# %% ../../notebooks/utils/00_basic.ipynb 20
@overload
def resample(
    df: pl.DataFrame,
    every: timedelta,
    period: timedelta = None,
    offset: timedelta = None,
    shift: timedelta = None,
    time_column="time",
)-> pl.DataFrame:
    ...

@overload
def resample(
    df: pl.LazyFrame,
    every: timedelta,
    period: timedelta = None,
    offset: timedelta = None,
    shift: timedelta = None,
    time_column="time",
)-> pl.LazyFrame:
    ...

def resample(
    df: pl.LazyFrame | pl.DataFrame,
    every: timedelta,
    period: timedelta = None,
    offset: timedelta = None,
    shift: timedelta = None,
    time_column="time",
):
    """Resample the DataFrame"""
    if period is None:
        period = every
    if shift is None:
        shift = period / 2
    return (
        df.sort(time_column)
        .group_by_dynamic(time_column, every=every, period=period, offset=offset)
        .agg(cs.numeric().mean())
        .with_columns((pl.col(time_column) + shift))
    )

# %% ../../notebooks/utils/00_basic.ipynb 21
def df2ts(
    df: Union[pandas.DataFrame, pl.DataFrame, pl.LazyFrame],
    cols = None,
    time_col = "time",
    attrs=None,
    name=None,
):
    """Convert DataFrame to TimeSeries"""
    
    if cols is None:
        if isinstance(df, (pl.DataFrame, pl.LazyFrame)):
            cols = df.columns
            cols.remove(time_col)
        else:
            cols = df.columns.tolist()
    
    if isinstance(df, pl.LazyFrame):
        df = df.collect()

    # Prepare data
    data = df[cols].to_numpy()

    # Prepare coordinates
    time = df.index if isinstance(df, pandas.DataFrame) else df[time_col]
    if isinstance(cols, str) and len(data.shape) > 1:
        element_len = data.shape[1]
        v_dim = [cols + str(i) for i in range(element_len)]
    else:
        v_dim = cols
        
    # Create the DataArray
    coords = {"time": time, "v_dim": v_dim}

    return xr.DataArray(data, coords=coords, attrs=attrs, name=name)


def calc_vec_mag(vec) -> DataArray:
    return linalg.norm(vec, dims="v_dim")

# %% ../../notebooks/utils/00_basic.ipynb 22
def check_fgm(vec: xr.DataArray):
    # check if time is monotonic increasing
    logger.info("Check if time is monotonic increasing")
    assert vec.time.to_series().is_monotonic_increasing
    # check available time difference
    logger.info(
        f"Available time delta: {vec.time.diff(dim='time').to_series().unique()}"
    )




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/utils/01_plotting.ipynb.

# %% auto 0
__all__ = ['savefig', 'time_stamp', 'setup_mva_plot', 'plot_candidate']

# %% ../../notebooks/utils/01_plotting.ipynb 1
import matplotlib.pyplot as plt

# %% ../../notebooks/utils/01_plotting.ipynb 2
def savefig(name, **kwargs):
    plt.savefig(f"../figures/{name}.png", bbox_inches="tight", **kwargs)
    plt.savefig(f"../figures/{name}.pdf", bbox_inches="tight", **kwargs)

# %% ../../notebooks/utils/01_plotting.ipynb 4
import xarray as xr
import pandas as pd
from datetime import datetime

from pyspedas.cotrans.minvar_matrix_make import minvar_matrix_make
from pyspedas import tvector_rotate
from pyspedas.analysis.tvectot import tvectot

from pytplot import tplot
from pytplot import store_data, get_data, split_vec, join_vec
from pytplot import timebar, highlight, degap, options


# %% ../../notebooks/utils/01_plotting.ipynb 5
import matplotlib.pyplot as plt
from .plot import savefig

try:
    import scienceplots
    plt.style.use(['science', 'nature', 'notebook'])
except ImportError:
    pass

# %% ../../notebooks/utils/01_plotting.ipynb 6
def time_stamp(ts):
    "Return POSIX timestamp as float."
    return pd.Timestamp(ts, tz="UTC").timestamp()


def setup_mva_plot(
    data: xr.DataArray,
    tstart: datetime,
    tstop: datetime,
    mva_tstart: datetime = None,
    mva_tstop: datetime = None,
):
    if mva_tstart is None:
        mva_tstart = tstart
    if mva_tstop is None:
        mva_tstop = tstop

    mva_b = data.sel(time=slice(mva_tstart, mva_tstop))
    store_data("fgm", data={"x": mva_b.time, "y": mva_b})
    minvar_matrix_make("fgm")  # get the MVA matrix

    temp_b = data.sel(time=slice(tstart, tstop))
    store_data("fgm", data={"x": temp_b.time, "y": temp_b})
    tvar = tvector_rotate("fgm_mva_mat", "fgm")[0]
    ysubtitle = "[nT LMN]"
    legend_names = [r"$B_l$", r"$B_m$", r"$B_n$"]
    
    tvar2plot = tvectot(tvar, join_component=True)
    legend_names = legend_names + [r"$B_{total}$"]

    options(tvar2plot, "ytitle", "$B$")
    options(tvar2plot, "ysubtitle", ysubtitle)
    options(tvar2plot, "legend_names", legend_names)
    
    options(tvar2plot, "thick", 2)
    options(tvar2plot, "char_size", 16)
    
    # tstart_ts = time_stamp(tstart)
    # tstop_ts = time_stamp(tstop)
    # highlight(tvar2plot, [tstart_ts, tstop_ts])
    degap(tvar2plot)
    return tvar2plot

# %% ../../notebooks/utils/01_plotting.ipynb 8
from matplotlib import lines


def plot_candidate(
    candidate: dict,
    data: xr.DataArray,
    add_timebars=True,
    add_plasma_params=False,
    plot_fit_data=False,
    **kwargs,
):
    if pd.notnull(candidate.get("d_tstart")) and pd.notnull(candidate.get("d_tstop")):
        tvar = setup_mva_plot(
            data,
            candidate["tstart"],
            candidate["tstop"],
            candidate["d_tstart"],
            candidate["d_tstop"],
        )
    else:
        tvar = setup_mva_plot(data, candidate["tstart"], candidate["tstop"])

    if add_timebars:
        d_time = candidate.get("d_time")
        d_start = candidate.get("d_tstart")
        d_stop = candidate.get("d_tstop")

        if d_time:
            timebar(time_stamp(d_time), color="red")
        if d_start and pd.notnull(d_start):
            timebar(time_stamp(d_start))
        if d_stop and pd.notnull(d_stop):
            timebar(time_stamp(d_stop))

    title = ""
    
    if add_plasma_params:
        plasma_speed = candidate.get("plasma_speed")
        plasma_density = candidate.get("plasma_density")
        plasma_temperature = candidate.get("plasma_temperature")

        title += "#Plasma parameters# "
        if plasma_speed:
            title += rf"$V_i: {plasma_speed:.2f} \mathrm{{km/s}}$"
        if plasma_density:
            title += rf", $n_i: {plasma_density:.2f} \mathrm{{cm}}^{{-3}}$"
        if plasma_temperature:
            title += rf", $T_i: {plasma_temperature:.2f} \mathrm{{eV}}$"

        # options(tvar, "title", title)

    fig, axes = tplot(tvar, return_plot_objects=True, **kwargs)

    if plot_fit_data:
        fit_data = candidate.get("fit.best_fit")
        fit_time = candidate.get("fit.time")

        c = candidate.get("fit.vars.c")
        amp = candidate.get("fit.vars.amplitude")
        
        d_star = candidate.get("d_star")
        rsquared = candidate.get("fit.stat.rsquared")
        chisqr = candidate.get("fit.stat.chisqr")

        plt.plot(d_time, c + amp / 2, marker="o", markersize=10, color="red")
        plt.plot(fit_time, fit_data, label="Fit", color="black", linestyle="--")
        
        title += f"\n#Fit# $\max dB/dt$: {d_star:.2f}, $R^2$: {rsquared:.2f}, $\chi^2$: {chisqr:.2f}"
    
    plt.title(title)

    return fig, axes



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/properties/00_mva.ipynb.

# %% auto 0
__all__ = ['minvar', 'calc_mva_features', 'fit_maxiumum_variance_direction', 'calc_candidate_mva_features']

# %% ../../notebooks/properties/00_mva.ipynb 1
import xarray as xr
import numpy as np
import pandas as pd
from datetime import timedelta

from lmfit.models import StepModel, ConstantModel, Model
from lmfit import Parameters

# %% ../../notebooks/properties/00_mva.ipynb 3
def minvar(data):
    """
    see `pyspedas.cotrans.minvar`

    This program computes the principal variance directions and variances of a
    vector quantity as well as the associated eigenvalues.

    Parameters
    -----------
    data:
        Vxyz, an (npoints, ndim) array of data(ie Nx3)

    Returns
    -------
    vrot:
        an array of (npoints, ndim) containing the rotated data in the new coordinate system, ijk.
        Vi(maximum direction)=vrot[0,:]
        Vj(intermediate direction)=vrot[1,:]
        Vk(minimum variance direction)=Vrot[2,:]
    v:
        an (ndim,ndim) array containing the principal axes vectors
        Maximum variance direction eigenvector, Vi=v[*,0]
        Intermediate variance direction, Vj=v[*,1] (descending order)
    w:
        the eigenvalues of the computation
    """

    #  Min var starts here
    # data must be Nx3
    vecavg = np.nanmean(np.nan_to_num(data, nan=0.0), axis=0)

    mvamat = np.zeros((3, 3))
    for i in range(3):
        for j in range(3):
            mvamat[i, j] = (
                np.nanmean(np.nan_to_num(data[:, i] * data[:, j], nan=0.0))
                - vecavg[i] * vecavg[j]
            )

    # Calculate eigenvalues and eigenvectors
    w, v = np.linalg.eigh(mvamat, UPLO="U")

    # Sorting to ensure descending order
    w = np.abs(w)
    idx = np.flip(np.argsort(w))

    # IDL compatability
    if True:
        if np.sum(w) == 0.0:
            idx = [0, 2, 1]

    w = w[idx]
    v = v[:, idx]

    # Rotate intermediate var direction if system is not Right Handed
    YcrossZdotX = v[0, 0] * (v[1, 1] * v[2, 2] - v[2, 1] * v[1, 2])
    if YcrossZdotX < 0:
        v[:, 1] = -v[:, 1]
        # v[:, 2] = -v[:, 2] # Should not it is being flipped at Z-axis?

    # Ensure minvar direction is along +Z (for FAC system)
    if v[2, 2] < 0:
        v[:, 2] = -v[:, 2]
        v[:, 1] = -v[:, 1]

    vrot = np.array([np.dot(row, v) for row in data])

    return vrot, v, w

# %% ../../notebooks/properties/00_mva.ipynb 5
def calc_mva_features(data: np.ndarray):
    """
    Compute MVA features based on the given data array.

    Parameters:
    - data (np.ndarray): Input data

    Returns:
    - List: Computed features
    """

    # Compute variance properties
    vrot, v, eigs = minvar(data)

    # Maximum variance direction eigenvector
    Vl = v[:, 0]
    Vn = v[:, 2]

    vec_mag = np.linalg.norm(vrot, axis=1)

    # Compute changes in each component of B_rot
    dvec = [vrot[0, i] - vrot[-1, i] for i in range(3)]

    # Compute mean values
    vec_mag_mean = np.mean(vec_mag)
    vec_n_mean = np.mean(vrot[:, 2])
    VnOverVmag = vec_n_mean / vec_mag_mean

    # Compute relative changes in magnitude
    dvec_mag = vec_mag[-1] - vec_mag[0]
    dBOverB = np.abs(dvec_mag / vec_mag_mean)
    dBOverB_max = (np.max(vec_mag) - np.min(vec_mag)) / vec_mag_mean

    output_names = [
        "Vl",
        "Vn",
        "b_mag",
        "b_n",
        "db_mag",
        "bn_over_b",
        "db_over_b",
        "db_over_b_max",
        "dB_lmn",
    ]

    results = [
        Vl,
        Vn,
        vec_mag_mean,
        vec_n_mean,
        dvec_mag,
        VnOverVmag,
        dBOverB,
        dBOverB_max,
        dvec,
    ]

    return pd.Series(results, index=output_names), vrot

# %% ../../notebooks/properties/00_mva.ipynb 8
def fit_maxiumum_variance_direction(
    ts: xr.DataArray, datetime_unit="s", return_best_fit: bool = True, **kwargs
):
    """
    Fit maximum variance direction data by model

    Note: 
        - see `datetime_to_numeric` in `xarray.core.duck_array_ops` for more details about converting datetime to numeric
        - Xarray uses the numpy dtypes datetime64[ns] and timedelta64[ns] to represent datetime data.
    """
    time = ts["time"].values
    x = (time - min(time)) / np.timedelta64(1, datetime_unit)
    y = ts.values

    xmin, xmax = min(x), max(x)
    ymin, ymax = min(y), max(y)

    # Create a model
    step_mod = StepModel(form="logistic")
    const_mod = ConstantModel()
    mod = step_mod + const_mod

    # Create parameters
    params = Parameters()
    params.add("c", value=0)
    params.add("center", value=(xmax + xmin) / 2.0, min=xmin, max=xmax)
    params.add("amplitude", value=(ymax - ymin))
    params.add("sigma", value=(xmax - xmin) / 7.0, min=0)

    out = mod.fit(y, params, x=x)

    amplitude = out.params["amplitude"].value
    sigma = out.params["sigma"].value
    center = out.params["center"].value
    max_df = amplitude / (4 * sigma)

    d_time = min(time) + center * np.timedelta64(1, datetime_unit).astype("timedelta64[ns]")

    result = pd.Series(
        {
            "fit.vars.amplitude": amplitude,
            "fit.vars.sigma": sigma,
            "d_time": d_time,
            "d_star": max_df,
            "fit.vars.c": out.params["c"].value,
            "fit.stat.rsquared": out.rsquared,
            "fit.stat.chisqr": out.chisqr,
        }
    )
    if return_best_fit:
        result["fit.best_fit"] = out.best_fit
        result["fit.time"] = time
    return result

# %% ../../notebooks/properties/00_mva.ipynb 9
def calc_candidate_mva_features(event, data: xr.DataArray, **kwargs):
    event_data = data.sel(time=slice(event["d_tstart"], event["d_tstop"]))

    mva_features, vrot = calc_mva_features(event_data.to_numpy())

    event_data_l = xr.DataArray(
        vrot[:, 0], dims=["time"], coords={"time": event_data.time}
    )

    fit_result = fit_maxiumum_variance_direction(event_data_l, **kwargs)

    return pd.concat([mva_features, fit_result])



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/ids_finder',
                'doc_host': 'https://Beforerr.github.io',
                'git_url': 'https://github.com/Beforerr/ids_finder',
                'lib_path': 'discontinuitypy'},
  'syms': { 'discontinuitypy.core.detection': { 'discontinuitypy.core.detection._compute_indices': ( 'ids_detection.html#_compute_indices',
                                                                                                     'discontinuitypy/core/detection.py'),
                                                'discontinuitypy.core.detection.add_neighbor_std': ( 'ids_detection.html#add_neighbor_std',
                                                                                                     'discontinuitypy/core/detection.py'),
                                                'discontinuitypy.core.detection.compute_combinded_std': ( 'ids_detection.html#compute_combinded_std',
                                                                                                          'discontinuitypy/core/detection.py'),
                                                'discontinuitypy.core.detection.compute_index_diff': ( 'ids_detection.html#compute_index_diff',
                                                                                                       'discontinuitypy/core/detection.py'),
                                                'discontinuitypy.core.detection.compute_index_fluctuation': ( 'ids_detection.html#compute_index_fluctuation',
                                                                                                              'discontinuitypy/core/detection.py'),
                                                'discontinuitypy.core.detection.compute_index_std': ( 'ids_detection.html#compute_index_std',
                                                                                                      'discontinuitypy/core/detection.py'),
                                                'discontinuitypy.core.detection.compute_indices': ( 'ids_detection.html#compute_indices',
                                                                                                    'discontinuitypy/core/detection.py'),
                                                'discontinuitypy.core.detection.compute_std': ( 'ids_detection.html#compute_std',
                                                                                                'discontinuitypy/core/detection.py'),
                                                'discontinuitypy.core.detection.detect_events': ( 'ids_detection.html#detect_events',
                                                                                                  'discontinuitypy/core/detection.py'),
                                                'discontinuitypy.core.detection.filter_indices': ( 'ids_detection.html#filter_indices',
                                                                                                   'discontinuitypy/core/detection.py'),
                                                'discontinuitypy.core.detection.pl_dvec': ( 'ids_detection.html#pl_dvec',
                                                                                            'discontinuitypy/core/detection.py'),
                                                'discontinuitypy.core.detection.pl_format_time': ( 'ids_detection.html#pl_format_time',
                                                                                                   'discontinuitypy/core/detection.py')},
            'discontinuitypy.core.event': { 'discontinuitypy.core.event.get_event_data': ( 'event.html#get_event_data',
                                                                                           'discontinuitypy/core/event.py')},
            'discontinuitypy.core.pipeline': { 'discontinuitypy.core.pipeline.compress_data_by_cands': ( 'ids_finder.html#compress_data_by_cands',
                                                                                                         'discontinuitypy/core/pipeline.py'),
                                               'discontinuitypy.core.pipeline.extract_features': ( 'ids_finder.html#extract_features',
                                                                                                   'discontinuitypy/core/pipeline.py'),
                                               'discontinuitypy.core.pipeline.ids_finder': ( 'ids_finder.html#ids_finder',
                                                                                             'discontinuitypy/core/pipeline.py')},
            'discontinuitypy.core.propeties': { 'discontinuitypy.core.propeties.IDsPipeline': ( 'ids_properties.html#idspipeline',
                                                                                                'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.IDsPipeline.__init__': ( 'ids_properties.html#idspipeline.__init__',
                                                                                                         'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.IDsPipeline.calc_duration': ( 'ids_properties.html#idspipeline.calc_duration',
                                                                                                              'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.IDsPipeline.calc_mva_features': ( 'ids_properties.html#idspipeline.calc_mva_features',
                                                                                                                  'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.IDsPipeline.calc_normal_direction': ( 'ids_properties.html#idspipeline.calc_normal_direction',
                                                                                                                      'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.IDsPipeline.calc_rotation_angle': ( 'ids_properties.html#idspipeline.calc_rotation_angle',
                                                                                                                    'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.IDsPipeline.calc_vec_change': ( 'ids_properties.html#idspipeline.calc_vec_change',
                                                                                                                'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.calc_candidate_duration': ( 'ids_properties.html#calc_candidate_duration',
                                                                                                            'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.calc_events_normal_direction': ( 'ids_properties.html#calc_events_normal_direction',
                                                                                                                 'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.calc_events_rotation_angle': ( 'ids_properties.html#calc_events_rotation_angle',
                                                                                                               'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.calc_events_vec_change': ( 'ids_properties.html#calc_events_vec_change',
                                                                                                           'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.calc_normal_direction': ( 'ids_properties.html#calc_normal_direction',
                                                                                                          'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.calc_rotation_angle': ( 'ids_properties.html#calc_rotation_angle',
                                                                                                        'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.get_candidate_data': ( 'ids_properties.html#get_candidate_data',
                                                                                                       'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.get_candidates': ( 'ids_properties.html#get_candidates',
                                                                                                   'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.get_data_at_times': ( 'ids_properties.html#get_data_at_times',
                                                                                                      'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.pdp.ApplyToRows._transform': ( 'ids_properties.html#pdp.applytorows._transform',
                                                                                                               'discontinuitypy/core/propeties.py'),
                                                'discontinuitypy.core.propeties.process_events': ( 'ids_properties.html#process_events',
                                                                                                   'discontinuitypy/core/propeties.py')},
            'discontinuitypy.datasets': { 'discontinuitypy.datasets.IDsDataset': ( 'datasets.html#idsdataset',
                                                                                   'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IDsDataset.Config': ( 'datasets.html#idsdataset.config',
                                                                                          'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IDsDataset.get_candidate': ( 'datasets.html#idsdataset.get_candidate',
                                                                                                 'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IDsDataset.get_candidate_data': ( 'datasets.html#idsdataset.get_candidate_data',
                                                                                                      'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IDsDataset.plot_candidate': ( 'datasets.html#idsdataset.plot_candidate',
                                                                                                  'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IDsDataset.plot_candidates': ( 'datasets.html#idsdataset.plot_candidates',
                                                                                                   'discontinuitypy/datasets.py'),
                                          'discontinuitypy.datasets.IDsDataset.update_candidates_with_plasma_data': ( 'datasets.html#idsdataset.update_candidates_with_plasma_data',
                                                                                                                      'discontinuitypy/datasets.py')},
            'discontinuitypy.datasets_kedro': { 'discontinuitypy.datasets_kedro.E1IDsDataset': ( 'datasets_kedro.html#e1idsdataset',
                                                                                                 'discontinuitypy/datasets_kedro.py'),
                                                'discontinuitypy.datasets_kedro.cIDsDataset': ( 'datasets_kedro.html#cidsdataset',
                                                                                                'discontinuitypy/datasets_kedro.py'),
                                                'discontinuitypy.datasets_kedro.cIDsDataset.__init__': ( 'datasets_kedro.html#cidsdataset.__init__',
                                                                                                         'discontinuitypy/datasets_kedro.py'),
                                                'discontinuitypy.datasets_kedro.cIDsDataset.load_data': ( 'datasets_kedro.html#cidsdataset.load_data',
                                                                                                          'discontinuitypy/datasets_kedro.py'),
                                                'discontinuitypy.datasets_kedro.cIDsDataset.load_events': ( 'datasets_kedro.html#cidsdataset.load_events',
                                                                                                            'discontinuitypy/datasets_kedro.py')},
            'discontinuitypy.integration': { 'discontinuitypy.integration.calc_combined_features': ( 'mag_plasma.html#calc_combined_features',
                                                                                                     'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.calc_rotation_angle_pl': ( 'mag_plasma.html#calc_rotation_angle_pl',
                                                                                                     'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.combine_features': ( 'mag_plasma.html#combine_features',
                                                                                               'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.compute_Alfven_current': ( 'mag_plasma.html#compute_alfven_current',
                                                                                                     'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.compute_Alfven_speed': ( 'mag_plasma.html#compute_alfven_speed',
                                                                                                   'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.compute_inertial_length': ( 'mag_plasma.html#compute_inertial_length',
                                                                                                      'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.vector_project': ( 'mag_plasma.html#vector_project',
                                                                                             'discontinuitypy/integration.py'),
                                             'discontinuitypy.integration.vector_project_pl': ( 'mag_plasma.html#vector_project_pl',
                                                                                                'discontinuitypy/integration.py')},
            'discontinuitypy.propeties.duration': { 'discontinuitypy.propeties.duration.calc_d_duration': ( 'properties/duration.html#calc_d_duration',
                                                                                                            'discontinuitypy/propeties/duration.py'),
                                                    'discontinuitypy.propeties.duration.calc_duration': ( 'properties/duration.html#calc_duration',
                                                                                                          'discontinuitypy/propeties/duration.py'),
                                                    'discontinuitypy.propeties.duration.find_start_end_times': ( 'properties/duration.html#find_start_end_times',
                                                                                                                 'discontinuitypy/propeties/duration.py'),
                                                    'discontinuitypy.propeties.duration.get_time_from_condition': ( 'properties/duration.html#get_time_from_condition',
                                                                                                                    'discontinuitypy/propeties/duration.py'),
                                                    'discontinuitypy.propeties.duration.ts_max_derivative': ( 'properties/duration.html#ts_max_derivative',
                                                                                                              'discontinuitypy/propeties/duration.py'),
                                                    'discontinuitypy.propeties.duration.ts_max_distance': ( 'properties/duration.html#ts_max_distance',
                                                                                                            'discontinuitypy/propeties/duration.py')},
            'discontinuitypy.propeties.mva': { 'discontinuitypy.propeties.mva.calc_candidate_mva_features': ( 'properties/mva.html#calc_candidate_mva_features',
                                                                                                              'discontinuitypy/propeties/mva.py'),
                                               'discontinuitypy.propeties.mva.calc_mva_features': ( 'properties/mva.html#calc_mva_features',
                                                                                                    'discontinuitypy/propeties/mva.py'),
                                               'discontinuitypy.propeties.mva.fit_maxiumum_variance_direction': ( 'properties/mva.html#fit_maxiumum_variance_direction',
                                                                                                                  'discontinuitypy/propeties/mva.py'),
                                               'discontinuitypy.propeties.mva.minvar': ( 'properties/mva.html#minvar',
                                                                                         'discontinuitypy/propeties/mva.py')},
            'discontinuitypy.utils.analysis': { 'discontinuitypy.utils.analysis.filter_before_jupiter': ( 'utils/analysis_utils.html#filter_before_jupiter',
                                                                                                          'discontinuitypy/utils/analysis.py'),
                                                'discontinuitypy.utils.analysis.filter_tranges_ds': ( 'utils/analysis_utils.html#filter_tranges_ds',
                                                                                                      'discontinuitypy/utils/analysis.py'),
                                                'discontinuitypy.utils.analysis.link_coord2dim': ( 'utils/analysis_utils.html#link_coord2dim',
                                                                                                   'discontinuitypy/utils/analysis.py'),
                                                'discontinuitypy.utils.analysis.n2_normalize': ( 'utils/analysis_utils.html#n2_normalize',
                                                                                                 'discontinuitypy/utils/analysis.py')},
            'discontinuitypy.utils.basic': { 'discontinuitypy.utils.basic.DataConfig': ( 'utils/basic.html#dataconfig',
                                                                                         'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic._expand_selectors': ( 'utils/basic.html#_expand_selectors',
                                                                                                'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.calc_vec_mag': ( 'utils/basic.html#calc_vec_mag',
                                                                                           'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.check_fgm': ( 'utils/basic.html#check_fgm',
                                                                                        'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.concat_df': ( 'utils/basic.html#concat_df',
                                                                                        'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.concat_partitions': ( 'utils/basic.html#concat_partitions',
                                                                                                'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.df2ts': ( 'utils/basic.html#df2ts',
                                                                                    'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.filter_tranges': ( 'utils/basic.html#filter_tranges',
                                                                                             'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.filter_tranges_df': ( 'utils/basic.html#filter_tranges_df',
                                                                                                'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.format_timedelta': ( 'utils/basic.html#format_timedelta',
                                                                                               'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.partition_data_by_time': ( 'utils/basic.html#partition_data_by_time',
                                                                                                     'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.partition_data_by_ts': ( 'utils/basic.html#partition_data_by_ts',
                                                                                                   'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.partition_data_by_year': ( 'utils/basic.html#partition_data_by_year',
                                                                                                     'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.partition_data_by_year_month': ( 'utils/basic.html#partition_data_by_year_month',
                                                                                                           'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.pl.DataFrame.plot': ( 'utils/basic.html#pl.dataframe.plot',
                                                                                                'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.pl_norm': ( 'utils/basic.html#pl_norm',
                                                                                      'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.pmap': ( 'utils/basic.html#pmap',
                                                                                   'discontinuitypy/utils/basic.py'),
                                             'discontinuitypy.utils.basic.resample': ( 'utils/basic.html#resample',
                                                                                       'discontinuitypy/utils/basic.py')},
            'discontinuitypy.utils.kedro': { 'discontinuitypy.utils.kedro.load_context': ( 'utils/kedro.html#load_context',
                                                                                           'discontinuitypy/utils/kedro.py')},
            'discontinuitypy.utils.lbl': { 'discontinuitypy.utils.lbl.LblDataset': ( 'utils/lbl.html#lbldataset',
                                                                                     'discontinuitypy/utils/lbl.py'),
                                           'discontinuitypy.utils.lbl.LblDataset.__init__': ( 'utils/lbl.html#lbldataset.__init__',
                                                                                              'discontinuitypy/utils/lbl.py'),
                                           'discontinuitypy.utils.lbl.LblDataset._describe': ( 'utils/lbl.html#lbldataset._describe',
                                                                                               'discontinuitypy/utils/lbl.py'),
                                           'discontinuitypy.utils.lbl.LblDataset._load': ( 'utils/lbl.html#lbldataset._load',
                                                                                           'discontinuitypy/utils/lbl.py'),
                                           'discontinuitypy.utils.lbl.LblDataset._save': ( 'utils/lbl.html#lbldataset._save',
                                                                                           'discontinuitypy/utils/lbl.py'),
                                           'discontinuitypy.utils.lbl.load_lbl': ( 'utils/lbl.html#load_lbl',
                                                                                   'discontinuitypy/utils/lbl.py')},
            'discontinuitypy.utils.plot': { 'discontinuitypy.utils.plot.plot_candidate': ( 'utils/plotting.html#plot_candidate',
                                                                                           'discontinuitypy/utils/plot.py'),
                                            'discontinuitypy.utils.plot.savefig': ( 'utils/plotting.html#savefig',
                                                                                    'discontinuitypy/utils/plot.py'),
                                            'discontinuitypy.utils.plot.setup_mva_plot': ( 'utils/plotting.html#setup_mva_plot',
                                                                                           'discontinuitypy/utils/plot.py'),
                                            'discontinuitypy.utils.plot.time_stamp': ( 'utils/plotting.html#time_stamp',
                                                                                       'discontinuitypy/utils/plot.py')},
            'discontinuitypy.utils.polars': { 'discontinuitypy.utils.polars._expand_selectors': ( 'utils/polars.html#_expand_selectors',
                                                                                                  'discontinuitypy/utils/polars.py'),
                                              'discontinuitypy.utils.polars.convert_to_pd_dataframe': ( 'utils/polars.html#convert_to_pd_dataframe',
                                                                                                        'discontinuitypy/utils/polars.py'),
                                              'discontinuitypy.utils.polars.create_partitions': ( 'utils/polars.html#create_partitions',
                                                                                                  'discontinuitypy/utils/polars.py'),
                                              'discontinuitypy.utils.polars.decompose_vector': ( 'utils/polars.html#decompose_vector',
                                                                                                 'discontinuitypy/utils/polars.py'),
                                              'discontinuitypy.utils.polars.pl_norm': ( 'utils/polars.html#pl_norm',
                                                                                        'discontinuitypy/utils/polars.py'),
                                              'discontinuitypy.utils.polars.sort': ( 'utils/polars.html#sort',
                                                                                     'discontinuitypy/utils/polars.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/utils/10_polars.ipynb.

# %% auto 0
__all__ = ['create_partitions', 'convert_to_pd_dataframe', 'sort', 'pl_norm', 'decompose_vector']

# %% ../../notebooks/utils/10_polars.ipynb 2
import polars as pl
import modin.pandas as mpd

from typing import Any, Collection

from functools import partial


# %% ../../notebooks/utils/10_polars.ipynb 3
def create_partitions(files, func):
    keys = [file.split("/")[-1] for file in files]
    return {key: partial(func, file) for key, file in zip(keys, files)}

# %% ../../notebooks/utils/10_polars.ipynb 5
def convert_to_pd_dataframe(
    df: pl.DataFrame | pl.LazyFrame, # original DataFrame or LazyFrame
    modin: bool = False # whether to use modin or not
):
    """
    Convert a Polars DataFrame or LazyFrame into a pandas-like DataFrame.
    If modin=True, returns a Modin DataFrame.
    """
    if isinstance(df, pl.LazyFrame):
        df = df.collect()
    elif not isinstance(df, pl.DataFrame):
        raise TypeError("Input must be a Polars DataFrame or LazyFrame")

    data = df.to_pandas(use_pyarrow_extension_array=True)

    if modin:
        return mpd.DataFrame(data)
    else:
        return data


# %% ../../notebooks/utils/10_polars.ipynb 7
def sort(df: pl.DataFrame, col="time"):
    if df.get_column(col).is_sorted():
        return df.set_sorted(col)
    else:
        return df.sort(col)


# %% ../../notebooks/utils/10_polars.ipynb 8
def _expand_selectors(items: Any, *more_items: Any) -> list[Any]:
    """
    See `_expand_selectors` in `polars`.
    """
    expanded: list[Any] = []
    for item in (
        *(
            items
            if isinstance(items, Collection) and not isinstance(items, str)
            else [items]
        ),
        *more_items,
    ):
        expanded.append(item)
    return expanded

# %% ../../notebooks/utils/10_polars.ipynb 9
def pl_norm(columns, *more_columns) -> pl.Expr:
    """
    Computes the square root of the sum of squares for the given columns.

    Args:
    *columns (str): Names of the columns.

    Returns:
    pl.Expr: Expression representing the square root of the sum of squares.
    """
    all_columns = _expand_selectors(columns, *more_columns)
    squares = [pl.col(column).pow(2) for column in all_columns]

    return sum(squares).sqrt()

# %% ../../notebooks/utils/10_polars.ipynb 10
def decompose_vector(df: pl.LazyFrame, vector_col, name=None):
    if name is None:
        name = vector_col

    return df.with_columns(
        pl.col(vector_col).list.get(0).alias(f"{name}_x"),
        pl.col(vector_col).list.get(1).alias(f"{name}_y"),
        pl.col(vector_col).list.get(2).alias(f"{name}_z"),
    )




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/21_event.ipynb.

# %% auto 0
__all__ = ['get_event_data']

# %% ../../notebooks/21_event.ipynb 1
import xarray as xr

# %% ../../notebooks/21_event.ipynb 3
def get_event_data(
    event: dict, data: xr.DataArray, neighbor: int = 0
) -> xr.DataArray:
    duration = event["tstop"] - event["tstart"]
    offset = neighbor * duration
    temp_tstart = event["tstart"] - offset
    temp_tstop = event["tstop"] + offset

    return data.sel(time=slice(temp_tstart, temp_tstop))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/utils/31_lbl.ipynb.

# %% auto 0
__all__ = ['HTTP_PROTOCOLS', 'load_lbl', 'LblDataset']

# %% ../../notebooks/utils/31_lbl.ipynb 2
from pathlib import PurePosixPath
import fsspec

from kedro.io import AbstractDataset
from kedro.io.core import get_filepath_str, get_protocol_and_path

import pandas
import pdr
import pooch

from typing import Any, Dict

# %% ../../notebooks/utils/31_lbl.ipynb 3
HTTP_PROTOCOLS = ("http", "https")


def load_lbl(filepath: str, type: str = "table") -> pandas.DataFrame:
    """Load LBL data.

    Args:
        filepath: File path to load the data from.
        type: Type of data to load. Options are 'table' and 'index'.

    Returns:
        A pandas DataFrame containing the loaded data.
    """
    if type == "table":
        df = pdr.read(filepath).TABLE
    elif type == "index":
        df = pandas.read_csv(filepath, delimiter=",", quotechar='"')
        df.columns = df.columns.str.replace(" ", "")

    return df


class LblDataset(AbstractDataset):
    def __init__(
        self,
        filepath: str,
        load_type: str = "table",
        metadata: Dict[str, Any] = None,
    ):
        # parse the path and protocol (e.g. file, http, s3, etc.)
        protocol, path = get_protocol_and_path(filepath)
        self._protocol = protocol
        self._filepath = PurePosixPath(path)

        self._fs = fsspec.filesystem(self._protocol)

        self.load_type = load_type
        self.metadata = metadata

    def _load(self):
        # using get_filepath_str ensures that the protocol and path are appended correctly for different filesystems
        load_path = get_filepath_str(self._filepath, self._protocol)

        if self._protocol in HTTP_PROTOCOLS:
            local_fp = pooch.retrieve(load_path, known_hash=None)
        else:
            local_fp = load_path

        return load_lbl(local_fp, self.load_type)

    def _save(self):
        pass

    def _describe(self):
        """Returns a dict that describes the attributes of the dataset."""
        return dict(filepath=self._filepath, protocol=self._protocol)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/__init__.ipynb.

# %% auto 0
__all__ = ['ROOT_DIR']

# %% ../notebooks/__init__.ipynb 1
__version__ = "0.0.1"

# %% ../notebooks/__init__.ipynb 2
from pathlib import Path

# %% ../notebooks/__init__.ipynb 3
#| eval: false
ROOT_DIR = Path(__file__).parent.parent

# %% ../notebooks/__init__.ipynb 6
# from discontinuitypy.utils.basic import load_params
# PARAMS = load_params()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/utils/15_kedro.ipynb.

# %% auto 0
__all__ = ['load_context']

# %% ../../notebooks/utils/15_kedro.ipynb 0
from kedro.config import OmegaConfigLoader
from kedro.io import DataCatalog

from kedro.framework.session import KedroSession
from kedro.framework.startup import bootstrap_project
from kedro.ipython import _resolve_project_path

# %% ../../notebooks/utils/15_kedro.ipynb 5
def load_context(project_path: str, params_only: bool = False, catalog_only: bool = False):
    project_path = _resolve_project_path(project_path)
    metadata = bootstrap_project(project_path)
    # configure_project(metadata.package_name)

    session = KedroSession.create(
        metadata.package_name, project_path,
    )
    context = session.load_context()

    if params_only:
        return context.params
    if catalog_only:
        return context.catalog
    else:
        return context



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/utils/02_analysis_utils.ipynb.

# %% auto 0
__all__ = ['AVG_SATS', 'link_coord2dim', 'filter_tranges_ds', 'filter_before_jupiter', 'n2_normalize']

# %% ../../notebooks/utils/02_analysis_utils.ipynb 2
import polars as pl
import pandas as pd

# %% ../../notebooks/utils/02_analysis_utils.ipynb 3
def link_coord2dim(df: pl.DataFrame, dim="time", coord: str = "radial_distance"):
    """Link the coord to a dimension across different subgroups

    Note: this idea is borrowed from the `xarray.DataArray.coords`.
    """
    base_df = df.filter(sat="JNO").select(dim, coord).rename({coord: f"ref_{coord}"})
    return df.join(base_df, on=dim, how="left")

# %% ../../notebooks/utils/02_analysis_utils.ipynb 4
from ..datasets import IDsDataset
from .basic import filter_tranges_df

# %% ../../notebooks/utils/02_analysis_utils.ipynb 5
def filter_tranges_ds(ds: IDsDataset, tranges: tuple[list, list]):
    """Filter a dataset by a list of time ranges"""
    new_ds = ds.copy()
    new_ds.candidates = filter_tranges_df(ds.candidates, tranges)
    new_ds.data = filter_tranges_df(ds.data.collect(), tranges).lazy()
    
    return new_ds

# %% ../../notebooks/utils/02_analysis_utils.ipynb 6
def filter_before_jupiter(df: pl.DataFrame):
    return df.filter(pl.col("time") < pd.Timestamp("2016-05-01"))

# %% ../../notebooks/utils/02_analysis_utils.ipynb 7
AVG_SATS = ["STA", "THB", "Wind"]


def n2_normalize(df: pl.DataFrame, cols, avg_sats: list = AVG_SATS):
    exprs = [pl.col(f"{col}").mean().alias(f"{col}_n2_factor") for col in cols]

    avg_df = df.filter(pl.col("sat").is_in(avg_sats)).group_by("time").agg(exprs)

    exprs = [
        (pl.col(f"{col}") / pl.col(f"{col}_n2_factor")).alias(f"{col}_n2")
        for col in cols
    ]

    return df.join(avg_df, on="time").with_columns(exprs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/20_datasets.ipynb.

# %% auto 0
__all__ = ['IDsDataset']

# %% ../notebooks/20_datasets.ipynb 1
import polars as pl
import pandas as pd
from loguru import logger
from random import sample

from pydantic import BaseModel, Field

# %% ../notebooks/20_datasets.ipynb 4
from .utils.basic import df2ts
from .utils.plot import plot_candidate as _plot_candidate
from .integration import combine_features, calc_combined_features

# %% ../notebooks/20_datasets.ipynb 5
class IDsDataset(BaseModel):
    class Config:
        arbitrary_types_allowed = True
        extra = "allow"

    mag_data: pl.LazyFrame = Field(alias="data")
    candidates: pl.DataFrame = None
    plasma_data: pl.LazyFrame = None

    bcols: list[str] = None
    vec_cols: list[str] = None

    def update_candidates_with_plasma_data(self):
        df_combined = combine_features(
            self.candidates, self.plasma_data.collect()
        ).lazy()

        self.candidates = calc_combined_features(
            df_combined, vec_cols=self.vec_cols
        ).collect()

    def get_candidate(self, index=None, predicates=None, **kwargs):
        if index is not None:
            candidate = self.candidates.row(index, named=True)
        elif predicates is not None:
            candidate = self.candidates.filter(predicates).row(0, named=True)
        return candidate

    def get_candidate_data(self, candidate=None, index=None, predicates=None, **kwargs):
        if candidate is None:
            candidate = self.get_candidate(index, predicates, **kwargs)

        _data = self.mag_data.filter(
            pl.col("time").is_between(candidate["tstart"], candidate["tstop"])
        )
        return df2ts(_data, self.bcols)

    def plot_candidate(self, candidate=None, index=None, predicates=None, **kwargs):
        if candidate is None:
            candidate = self.get_candidate(index, predicates, **kwargs)
        sat_fgm = self.get_candidate_data(candidate)

        return _plot_candidate(candidate, sat_fgm, **kwargs)

    def plot_candidates(self, indices=None, num=4, random=True, **kwargs):
        if indices is None:
            indices = sample(range(len(self.candidates)), num) if random else range(num)
            logger.info(f"Candidates indices: {indices}")

        return [self.plot_candidate(index=i, **kwargs) for i in indices]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/01_ids_detection.ipynb.

# %% auto 0
__all__ = ['INDEX_STD_THRESHOLD', 'INDEX_FLUC_THRESHOLD', 'INDEX_DIFF_THRESHOLD', 'SPARSE_THRESHOLD', 'pl_format_time',
           'compute_std', 'compute_combinded_std', 'add_neighbor_std', 'compute_index_std', 'compute_index_fluctuation',
           'pl_dvec', 'compute_index_diff', 'compute_indices', 'filter_indices', 'detect_events']

# %% ../../notebooks/01_ids_detection.ipynb 3
from datetime import timedelta
import polars as pl

from fastcore.utils import *
from fastcore.test import *
from ..utils.polars import pl_norm

# %% ../../notebooks/01_ids_detection.ipynb 4
from ..utils.basic import _expand_selectors

# %% ../../notebooks/01_ids_detection.ipynb 5
# some helper functions
def pl_format_time(df: pl.LazyFrame | pl.DataFrame, tau: timedelta):
    return df.with_columns(
        tstart=pl.col("time"),
        tstop=(pl.col("time") + tau),
        time=(pl.col("time") + tau / 2),
    )

# %% ../../notebooks/01_ids_detection.ipynb 6
def compute_std(
    df: pl.DataFrame | pl.LazyFrame,
    period: timedelta,  # period to group by
    b_cols=["BX", "BY", "BZ"],
    every: timedelta = None,  # every to group by (default: period / 2)
):
    if every is None:
        every = period / 2
    b_std_cols = [col_name + "_std" for col_name in b_cols]

    std_df = (
        df.group_by_dynamic("time", every=every, period=period)
        .agg(
            pl.count(),
            pl.col(b_cols).std(ddof=0).map_alias(lambda col_name: col_name + "_std"),
        )
        .with_columns(
            pl_norm(b_std_cols).alias("B_std"),
        )
        .drop(b_std_cols)
    )
    return std_df


def compute_combinded_std(df: pl.DataFrame | pl.LazyFrame, tau, cols) -> pl.DataFrame:
    combined_std_cols = [col_name + "_combined_std" for col_name in cols]
    offsets = [0 * tau, tau / 2]
    combined_std_dfs = []
    for offset in offsets:
        truncated_df = df.select(
            (pl.col("time") - offset).dt.truncate(tau, offset=offset).alias("time"),
            pl.col(cols),
        )

        prev_df = truncated_df.select(
            (pl.col("time") + tau),
            pl.col(cols),
        )

        next_df = truncated_df.select(
            (pl.col("time") - tau),
            pl.col(cols),
        )

        temp_combined_std_df = (
            pl.concat([prev_df, next_df])
            .group_by("time")
            .agg(
                pl.col(cols)
                .std(ddof=0)
                .map_alias(lambda col_name: col_name + "_combined_std"),
            )
            .with_columns(B_std_combined=pl_norm(combined_std_cols))
            .drop(combined_std_cols)
            .sort("time")
        )

        combined_std_dfs.append(temp_combined_std_df)

    combined_std_df = pl.concat(combined_std_dfs)
    return combined_std_df

# %% ../../notebooks/01_ids_detection.ipynb 8
def add_neighbor_std(
    df: pl.LazyFrame, tau: timedelta, join_strategy="inner"
):  # noqa: F811
    """
    Get the neighbor standard deviations

    Parameters
    ----------
    - df (pl.LazyFrame): The input DataFrame.
    - tau : The time interval value.

    Notes
    -----
    Simply shift would not work correctly if data is missing, like `std_next = pl.col("B_std").shift(-2)`.

    """

    # Calculate the standard deviation index
    prev_std_df = df.select(
        pl.col("time") + tau,
        B_std_prev=pl.col("B_std"),
        count_prev=pl.col("count"),
    )

    next_std_df = df.select(
        pl.col("time") - tau,
        B_std_next=pl.col("B_std"),
        count_next=pl.col("count"),
    )

    return df.join(prev_std_df, on="time", how=join_strategy).join(
        next_std_df, on="time", how=join_strategy
    )

# %% ../../notebooks/01_ids_detection.ipynb 9
def compute_index_std(df: pl.LazyFrame):  # noqa: F811
    """
    Compute the standard deviation index based on the given DataFrame

    Parameters
    ----------
    - df (pl.LazyFrame): The input DataFrame.

    Returns
    -------
    - pl.LazyFrame: DataFrame with calculated 'index_std' column.

    Examples
    --------
    >>> index_std_df = compute_index_std_pl(df)
    >>> index_std_df
    """

    return df.with_columns(
        index_std=pl.col("B_std") / pl.max_horizontal("B_std_prev", "B_std_next"),
    )

# %% ../../notebooks/01_ids_detection.ipynb 11
def compute_index_fluctuation(df: pl.LazyFrame, base_col="B_std"):
    std_combined = pl.col("B_std_combined")
    std_added = pl.sum_horizontal("B_std_prev", "B_std_next")
    return df.with_columns(index_fluctuation=std_combined / std_added)

# %% ../../notebooks/01_ids_detection.ipynb 13
def pl_dvec(columns, *more_columns):
    all_columns = _expand_selectors(columns, *more_columns)
    return [
        (pl.col(column).first() - pl.col(column).last()).alias(f"d{column}_vec")
        for column in all_columns
    ]

# %% ../../notebooks/01_ids_detection.ipynb 14
def compute_index_diff(df: pl.LazyFrame, period: timedelta, cols):
    db_cols = ["d" + col + "_vec" for col in cols]

    index_diff = (
        df.with_columns(B=pl_norm(cols))
        .group_by_dynamic("time", every=period / 2, period=period)
        .agg(
            pl.col("B").mean().alias("B_mean"),
            *pl_dvec(cols),
        )
        .with_columns(dB_vec=pl_norm(db_cols))
        .with_columns(index_diff=pl.col("dB_vec") / pl.col("B_mean"))
        .drop(db_cols)
    )

    return index_diff

# %% ../../notebooks/01_ids_detection.ipynb 15
def _compute_indices(
    df: pl.LazyFrame, tau: timedelta, cols: list[str] = ["BX", "BY", "BZ"]
) -> pl.LazyFrame:
    join_strategy = "inner"

    std_df = compute_std(df, tau, cols)
    stds_df = add_neighbor_std(std_df, tau)

    combined_std_df = compute_combinded_std(df, tau, cols)
    index_diff = compute_index_diff(df, tau, cols)

    indices = (
        stds_df.join(index_diff, on="time")
        .join(combined_std_df, on="time", how=join_strategy)
        .pipe(compute_index_std)
        .pipe(compute_index_fluctuation)
        .drop(["B_std_prev", "B_std_next", "B_added_std", "B_std_combined"])
    )

    return indices


def compute_indices(
    df: pl.DataFrame, tau: timedelta, bcols: list[str] = ["BX", "BY", "BZ"]
) -> pl.DataFrame:
    """
    Compute all index based on the given DataFrame and tau value.

    Parameters
    ----------
    df : pl.DataFrame
        Input DataFrame.
    tau : datetime.timedelta
        Time interval value.

    Returns
    -------
    tuple :
        Tuple containing DataFrame results for fluctuation index,
        standard deviation index, and 'index_num'.

    Examples
    --------
    >>> indices = compute_indices(df, tau)

    Notes
    -----
    - This is a wrapper for `_compute_indices` with `pl.LazyFrame` input.
    - Simply shift to calculate index_std would not work correctly if data is missing,
        like `std_next = pl.col("B_std").shift(-2)`.
    - Drop null though may lose some IDs (using the default `join_strategy`).
        Because we could not tell if it is a real ID or just a partial wave
        from incomplete data without previous or/and next std.
        Hopefully we can pick up the lost ones with smaller tau.
    - TODO: Can be optimized further, but this is already fast enough.
        - TEST: if `join` can be improved by shift after filling the missing values.
        - TEST: if `list` in `polars` really fast?
    """
    return _compute_indices(df.lazy(), tau, bcols).collect()

# %% ../../notebooks/01_ids_detection.ipynb 17
# from discontinuitypy import PARAMS

# INDEX_STD_THRESHOLD = PARAMS['detection']['index_std_threshold']
# INDEX_FLUC_THRESHOLD = PARAMS['detection']['index_fluc_threshold']
# INDEX_DIFF_THRESHOLD = PARAMS['detection']['index_diff_threshold']
INDEX_STD_THRESHOLD = 2
INDEX_FLUC_THRESHOLD = 1
INDEX_DIFF_THRESHOLD = 0.1
SPARSE_THRESHOLD = 15

# %% ../../notebooks/01_ids_detection.ipynb 18
def filter_indices(
    df: pl.DataFrame | pl.LazyFrame,
    index_std_threshold : float = INDEX_STD_THRESHOLD,
    index_fluc_threshold : float =INDEX_FLUC_THRESHOLD,
    index_diff_threshold : float =INDEX_DIFF_THRESHOLD,
    sparse_num : int = SPARSE_THRESHOLD,
) -> pl.DataFrame | pl.LazyFrame:
    # filter indices to get possible IDs

    return df.filter(
        pl.col("index_std") > index_std_threshold,
        pl.col("index_fluctuation") > index_fluc_threshold,
        pl.col("index_diff") > index_diff_threshold,
        pl.col("index_std").is_finite(), # for cases where neighboring groups have std=0
        pl.col("count") > sparse_num, 
        pl.col("count_prev") > sparse_num, # filter out sparse intervals, which may give unreasonable results.
        pl.col("count_next") > sparse_num # filter out sparse intervals, which may give unreasonable results.
    ).drop(["count_prev", "count_next"])

# %% ../../notebooks/01_ids_detection.ipynb 19
def detect_events(data: pl.DataFrame, tau: timedelta, ts: timedelta, bcols):
    indices = compute_indices(data, tau, bcols)
    sparse_num = tau / ts // 3
    events = indices.pipe(filter_indices, sparse_num=sparse_num).pipe(
        pl_format_time, tau
    )
    return events



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/00_ids_finder.ipynb.

# %% auto 0
__all__ = ['compress_data_by_cands', 'ids_finder', 'extract_features']

# %% ../../notebooks/00_ids_finder.ipynb 2
#| code-summary: "Import all the packages needed for the project"
import polars as pl
from .propeties import process_events
from .detection import detect_events
from ..utils.basic import df2ts
import numpy as np

from datetime import timedelta

from typing import Callable

# %% ../../notebooks/00_ids_finder.ipynb 5
def compress_data_by_cands(
    data: pl.DataFrame, candidates: pl.DataFrame
):
    """Compress the data for parallel processing"""
    ttstarts = candidates["tstart"]
    ttstops = candidates["tstop"]

    ttstarts_index = data["time"].search_sorted(ttstarts)
    ttstops_index = data["time"].search_sorted(ttstops)

    indices = np.concatenate(
        [
            np.arange(ttstart_index, ttstop_index + 1)
            for ttstart_index, ttstop_index in zip(ttstarts_index, ttstops_index)
        ]
    )  # faster than `pl.arange`
    indices_unique = (
        pl.Series(indices).unique().sort()
    )  # faster than `np.unique(index)`
    return data[indices_unique]


# %% ../../notebooks/00_ids_finder.ipynb 6
def ids_finder(
    detection_df: pl.LazyFrame, # data used for anomaly dectection (typically low cadence data)
    tau: timedelta,
    ts: timedelta, 
    bcols = None,
    extract_df: pl.LazyFrame = None, # data used for feature extraction (typically high cadence data)
):
    if extract_df is None:
        extract_df = detection_df
    if bcols is None:
        bcols = detection_df.columns
        bcols.remove("time")
    
    detection_df = detection_df.sort("time").with_columns(pl.col("time").dt.cast_time_unit("us")) # https://github.com/pola-rs/polars/issues/12023
    extract_df = extract_df.sort("time").with_columns(pl.col("time").dt.cast_time_unit("us"))

    events = detect_events(detection_df, tau, ts, bcols)
    
    data_c = compress_data_by_cands(extract_df.collect(), events)
    sat_fgm = df2ts(data_c, bcols)
    ids = process_events(events, sat_fgm, ts)
    return ids

# %% ../../notebooks/00_ids_finder.ipynb 8
def extract_features(
    partitioned_input: dict[str, Callable[..., pl.LazyFrame]],
    tau: float,  # in seconds, yaml input
    ts: float,  # in seconds, yaml input
    **kwargs,
) -> pl.DataFrame:
    "wrapper function for partitioned input"

    _tau = timedelta(seconds=tau)
    _ts = timedelta(seconds=ts)

    ids = pl.concat(
        [
            ids_finder(partition_load(), _tau, _ts, **kwargs)
            for partition_load in partitioned_input.values()
        ]
    )
    return ids.unique(["d_time", "d_tstart", "d_tstop"])



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/03_mag_plasma.ipynb.

# %% auto 0
__all__ = ['combine_features', 'vector_project', 'vector_project_pl', 'calc_rotation_angle_pl', 'compute_inertial_length',
           'compute_Alfven_speed', 'compute_Alfven_current', 'calc_combined_features']

# %% ../notebooks/03_mag_plasma.ipynb 1
import polars as pl
import polars.selectors as cs

# %% ../notebooks/03_mag_plasma.ipynb 2
def combine_features(candidates: pl.DataFrame, states_data: pl.DataFrame):
    # change time format: see issue: https://github.com/pola-rs/polars/issues/12023
    states_data = states_data.with_columns(
        cs.datetime().dt.cast_time_unit("ns"),
    )
    candidates = candidates.with_columns(
        cs.datetime().dt.cast_time_unit("ns"),
    )

    return candidates.sort("time").join_asof(states_data.sort("time"), on="time")

# %% ../notebooks/03_mag_plasma.ipynb 4
import astropy.units as u
from astropy.constants import mu0, e
from plasmapy.formulary.lengths import inertial_length
from plasmapy.formulary.speeds import Alfven_speed
from xarray_einstats import linalg
from .utils.basic import df2ts
from .core.propeties import calc_rotation_angle
import xarray as xr

# %% ../notebooks/03_mag_plasma.ipynb 5
def vector_project(v1,v2, dim="v_dim"):
    return xr.dot(v1 , v2, dims=dim) / linalg.norm(v2, dims=dim)

def vector_project_pl(df: pl.LazyFrame, v1_cols, v2_cols, name=None):
    
    v1 = df2ts(df, v1_cols).assign_coords(v_dim=["x","y","z"])
    v2 = df2ts(df, v2_cols).assign_coords(v_dim=["x","y","z"])  
    result = vector_project(v1, v2, dim="v_dim")
    
    return df.with_columns(
        pl.Series(result.data).alias(name or "v_proj")
    )

# %% ../notebooks/03_mag_plasma.ipynb 6
def calc_rotation_angle_pl(ldf: pl.LazyFrame, v1_cols, v2_cols, name):
    df = ldf.collect()
    v1 = df.select(v1_cols).to_numpy()
    v2 = df.select(v2_cols).to_numpy()
    
    result = calc_rotation_angle(v1, v2)
    
    return ldf.with_columns(
        pl.Series(result).alias(name)
    )

# %% ../notebooks/03_mag_plasma.ipynb 8
def compute_inertial_length(ldf: pl.LazyFrame, density_col = "plasma_density"):
    df = ldf.collect()

    density = df[density_col].to_numpy() * u.cm ** (-3)
    result = inertial_length(density, "H+").to(u.km)

    return df.with_columns(ion_inertial_length=pl.Series(result.value)).lazy()

# %% ../notebooks/03_mag_plasma.ipynb 10
def compute_Alfven_speed(
    ldf: pl.LazyFrame,
    density_col = "plasma_density",
):
    df = ldf.collect()

    B = df["B"] if "B" in df.columns else df["b_mag"]  # backwards compatiblity
    density = df[density_col].to_numpy() * u.cm ** (-3)
    result = Alfven_speed(B.to_numpy() * u.nT, density=density, ion="p+").to(u.km / u.s)

    return df.with_columns(Alfven_speed=pl.Series(result.value)).lazy()


def compute_Alfven_current(
    ldf: pl.LazyFrame,
    density_col = "plasma_density",
):
    df = ldf.collect()

    Alfven_speed = df["Alfven_speed"].to_numpy() * u.km / u.s
    density = df[density_col].to_numpy() * u.cm ** (-3)

    result = (e.si * Alfven_speed * density)
    result = result.to(u.nA / u.m**2)

    return df.with_columns(j_Alfven=pl.Series(result.value)).lazy()

# %% ../notebooks/03_mag_plasma.ipynb 12
def calc_combined_features(
    df: pl.LazyFrame,
    vec_cols = ["v_x", "v_y", "v_z"],  # plasma velocity vector in any fixed coordinate system,
    normal_cols = ["k_x", "k_y", "k_z"], # normal vector of the discontinuity plane
    b_cols = None, # ["B_background_x", "B_background_y", "B_background_z"]
):
    
    Vl_cols = ["Vl_x", "Vl_y", "Vl_z"]
    Vn_cols = ["Vn_x", "Vn_y", "Vn_z"]
    
    j_factor = ((u.nT / u.s) * (1 / mu0 / (u.km / u.s))).to(u.nA / u.m**2)

    result = df.with_columns( duration = pl.col("d_tstop") - pl.col("d_tstart"))

    if b_cols is not None:
        result = result.pipe(calc_rotation_angle_pl, b_cols, normal_cols, name="theta_n_b")
        
    result = (
        result
        .pipe(vector_project_pl, vec_cols, Vl_cols, name="v_l")
        .pipe(vector_project_pl, vec_cols, Vn_cols, name="v_n")
        .pipe(vector_project_pl, vec_cols, normal_cols, name="v_k")
        .with_columns(
            pl.col("v_n").abs(),
            pl.col("v_k").abs(),
            v_mn=(pl.col("plasma_speed") ** 2 - pl.col("v_l") ** 2).sqrt(),
        )
        .with_columns(
            L_n=pl.col("v_n") * pl.col("duration").dt.nanoseconds() / 1e9,
            L_mn=pl.col("v_mn") * pl.col("duration").dt.nanoseconds() / 1e9,
            L_k=pl.col("v_k") * pl.col("duration").dt.nanoseconds() / 1e9,
            j0=pl.col("d_star") / pl.col("v_mn"),
            j0_k=pl.col("d_star") / pl.col("v_k"),
        )
        .pipe(compute_inertial_length)
        .pipe(compute_Alfven_speed)
        .pipe(compute_Alfven_current)
        .with_columns(
            cs.by_name("j0", "j0_k") * j_factor.value,
        )
    )
    return result
